#!/usr/bin/env python


from contextlib import nullcontext as no_error
from typing import Generator, Mapping, Optional, Union

import pytest
from _pytest.python_api import RaisesContext
from hatch.utils.fs import Path
from pytest import raises

from hatch_semver.bump_instruction import BumpInstruction as BI
from hatch_semver.errors import ValidationError
from hatch_semver.semver_scheme import SemverScheme

# define some shorthand variables for writing the test parameters
sep = SemverScheme.INSTRUCTION_SEPARATOR.join
bsep = BI.sep.join

val = {"validate-bump": True}
no_val = {"validate-bump": False}

ok = no_error()
not_higher = raises(ValidationError, match="not higher")
not_as_high = raises(ValidationError, match="at least as high")
directly = raises(ValueError, match="directly")
specifically = raises(ValueError, match="specifically")
invalid = raises(ValueError, match="not valid SemVer")


@pytest.mark.parametrize(
    "instructions, original, expected, settings, exp_error",
    (
        # no instructions, just give us the current version
        ("", "0.0.1", "0.0.1", val, ok),
        ("", "0.0.1", "0.0.1", no_val, ok),
        ("", "0.0.1", "0.0.1", {}, ok),
        (None, "0.1.0", "0.1.0", val, ok),
        (None, "0.1.0", "0.1.0", no_val, ok),
        (None, "0.1.0", "0.1.0", {}, ok),
        # bump build
        ("build", "2.3.4", "2.3.4+build.1", val, ok),
        ("build", "2.3.4+build.1", "2.3.4+build.2", no_val, ok),
        ("build", "2.3.4+build.1", "2.3.4+build.2", val, ok),
        (sep(("minor", "build")), "2.3.4+build.1", "2.4.0+build.1", val, ok),
        ("2.3.4+other-build.1", "2.3.4+build.1", "2.3.4+other-build.1", val, ok),
        ("2.3.4+other-build.1", "2.3.4+build.1", "2.3.4+other-build.1", no_val, ok),
        # custom unnumbered build cannot be bumped
        ("build", "2.3.4+arst", "2.3.4+arst", {}, ok),
        # custom numbered build can be bumped
        ("build", "2.3.4+arst.0", "2.3.4+arst.1", {}, ok),
        ("build", "2.3.4+arst.1", "2.3.4+arst.2", {}, ok),
        ("build", "2.3.4+devdrop0", "2.3.4+devdrop1", {}, ok),
        ("build", "4.3.3-beta.4+custom.3", "4.3.3-beta.4+custom.4", {}, ok),
        ("build", "4.3.3-rc.1", "4.3.3-rc.1+build.1", {}, ok),
        ("build=dev", "2.9.3", "2.9.3+dev.1", val, ok),
        ("dev", "2.9.3", "2.9.3+dev.1", val, ok),
        ("dev=develop", "2.9.3", "2.9.3+develop.1", val, directly),
        # bump to release
        ("release", "2.3.4-beta.3+devdrop0", "2.3.4", {}, ok),
        ("release", "2.3.4-beta.3+devdrop0", "2.3.4", val, ok),
        ("release", "2.3.4-beta.3+devdrop0", "2.3.4", no_val, ok),
        ("release", "12.35.49-rc.3", "12.35.49", {}, ok),
        ("release", "6.4.2-custom-pre-release", "6.4.2", {}, ok),
        ("release", "6.4.2-custom-pre-release", "6.4.2", val, ok),
        ("release", "6.4.2-custom-pre-release", "6.4.2", no_val, ok),
        ("release", "12.35.49+build.3", "12.35.49", no_val, ok),
        ("release", "5.2.3", "5.2.3", no_val, ok),
        ("release", "2.3.4", "2.3.4", {}, not_higher),
        ("release", "2.3.4", "2.3.4", val, not_higher),
        ("release", "12.35.49+build.3", "12.35.49", val, not_higher),
        ("release,beta", "7.1.8-rc.1", "7.1.9-beta.1", val, ok),
        # specific version
        ("3.21.7-gamma+didaktik", "1.0.1", "3.21.7-gamma+didaktik", {}, ok),
        ("3.21.7-gamma+didaktik", "1.0.1", "3.21.7-gamma+didaktik", val, ok),
        ("3.21.7-gamma+didaktik", "1.0.1", "3.21.7-gamma+didaktik", no_val, ok),
        ("3.21.7-gamma+didaktik", "3.21.7-gandalf", "3.21.7-gamma+didaktik", {}, not_higher),
        ("3.21.7-gamma+didaktik", "3.21.7-gandalf", "3.21.7-gamma+didaktik", val, not_higher),
        ("3.21.7-gamma+didaktik", "3.21.7-gandalf", "3.21.7-gamma+didaktik", no_val, ok),
        ("7.34.2-alpha+data", "7.34.2-alpha.1", "7.34.2-alpha+data", {}, not_higher),
        ("7.34.2-alpha.1+data.1", "7.34.2-alpha.2", "7.34.2-alpha.1+data.1", {}, not_higher),
        ("7.34.3-alpha.1+data.1", "7.34.2", "7.34.3-alpha.1+data.1", {}, ok),
        ("7.34.2-alpha.1+data.1", "7.34.2-alpha.1", "7.34.2-alpha.1+data.1", {}, ok),
        ("7.34.2-alpha.1+data.1", "7.34.2-alpha.1", "7.34.2-alpha.1+data.1", no_val, ok),
        (bsep(("build", "data")), "7.34.2-alpha.1", "7.34.2-alpha.1+data.1", {}, ok),
        ("3.0.0-almost-done.3", "3.0.0-beta.2", "3.0.0-almost-done.3", {}, not_higher),
        ("3.0.0-almost-done.3", "3.0.0-beta.2+build.1", "3.0.0-almost-done.3", {}, not_higher),
        # change build metadata by entering a specific version
        ("9.1.2-alpha.3+dev.1", "9.1.2-alpha.3+build.1", "9.1.2-alpha.3+dev.1", {}, ok),
        # attempt to bump an alphanumeric identifier (python-semver behavior)
        ("build", "6.3.4-rc.2+verbose", "6.3.4-rc.2+verbose", {}, ok),
        ("build", "6.3.4-rc.2+verbose", "6.3.4-rc.2+verbose", val, ok),
        ("build", "6.3.4-rc.2+verbose", "6.3.4-rc.2+verbose", no_val, ok),
        # multiple instructions
        (sep(("1.0.0", "build")), "1.1.0", "1.0.0+build.1", val, not_as_high),
        (sep(("1.0.0", "build")), "1.1.0", "1.0.0+build.1", {}, not_as_high),
        (sep(("alpha", bsep(("build", "data")))), "7.34.2", "7.34.3-alpha.1+data.1", {}, ok),
        (sep((bsep(("build", "data")), "alpha")), "7.34.2", "7.34.3-alpha.1", {}, ok),
        (sep(("1.0.0", "build")), "1.1.0", "1.0.0+build.1", no_val, ok),
        # temporary precedence violations in intermediate versions
        (sep(("major", "8.0.5")), "8.0.4", "8.0.5", val, ok),
        (sep(("8.0.3", "patch", "build")), "8.0.4", "8.0.4+build.1", val, ok),
        (sep(("8.0.3", "patch", "build")), "8.0.4+build.1", "8.0.4+build.1", val, ok),
        (sep(("8.0.4", "alpha", "pre")), "8.0.4-alpha.1", "8.0.4-alpha.2", val, ok),
        # questionable custom value
        (bsep(("build", "build.1")), "9.8.4", "9.8.4+build.1.1", {}, ok),
        # wrong custom values
        (bsep(("alpha", "myalpha")), "1.3.5", "", {}, directly),
        (bsep(("beta", "mybeta")), "1.3.5", "", val, directly),
        (bsep(("minor", "4")), "1.3.5", "", no_val, specifically),
        # release candidates
        ("rc", "4.2.1", "4.2.2-rc.1", {}, ok),
        ("pre", "4.2.1", "4.2.2-rc.1", {}, ok),
        ("prerelease", "4.2.1", "4.2.2-rc.1", {}, ok),
        ("pre-release", "4.2.1", "4.2.2-rc.1", {}, ok),
        (bsep(("rc", "myrc")), "1.3.5", "1.3.6-myrc.1", val, ok),
        (sep(("rc", "rc", "rc", "rc")), "6.4.3", "6.4.4-rc.4", {}, ok),
        # prereleases
        ("alpha", "1.0.0+dev", "1.0.1-alpha.1", {}, ok),
        ("beta", "1.0.0+dev", "1.0.1-beta.1", {}, ok),
        (bsep(("pre", "gamma")), "1.0.0+dev", "1.0.1-gamma.1", {}, ok),
        ("rc=preview", "3.2.1", "3.2.2-preview.1", {}, ok),
        ("rc", "3.2.1-preview", "3.2.1-preview", {}, not_higher),
        # custom prerelease string once set, cannot be reset,
        # that's a python-semver bug https://github.com/python-semver/python-semver/issues/339
        ("prerelease=beta", "3.2.1-alpha.1", "3.2.1-alpha.2", {}, ok),
        # chained prereleases
        (sep(("patch", "rc")), "1.0.0", "1.0.1-rc.1", {}, ok),
        (sep(("minor", "rc")), "2.3.8", "2.4.0-rc.1", {}, ok),
        (sep(("major", bsep(("rc", "alpha")))), "0.9.5", "1.0.0-alpha.1", {}, ok),
        (sep(("major", "rc", "build")), "0.5.0", "1.0.0-rc.1+build.1", {}, ok),
        (sep(("major", "build", "rc")), "0.5.0", "1.0.0-rc.1", {}, ok),
        (sep(("1.5.0", "rc")), "2.0.0", "1.5.0-rc.1", {}, not_higher),
        (sep(("1.5.0", "rc")), "2.0.0", "1.5.0-rc.1", val, not_higher),
        (sep(("1.5.0", "rc")), "2.0.0", "1.5.0-rc.1", no_val, ok),
        (sep(("2.5.0", "rc")), "2.0.0", "2.5.0-rc.1", {}, ok),
        (sep(("2.5.0+build.1", "rc")), "2.0.0", "2.5.0-rc.1", {}, ok),
        (sep(("2.5.0", "build", "rc")), "2.0.0", "2.5.0-rc.1", {}, ok),
        (sep(("2.5.0", "rc", "build")), "2.0.0", "2.5.0-rc.1+build.1", {}, ok),
        (sep(("minor", "rc", "rc")), "7.0.23", "7.1.0-rc.2", {}, ok),
        (sep(("minor", "rc", "build", "rc")), "7.0.23", "7.1.0-rc.2", {}, ok),
        ("rc", "7.1.0-rc.1+build.1", "7.1.0-rc.2", {}, ok),
        # custom prerelease string once set, cannot be reset,
        # that's a python-semver bug https://github.com/python-semver/python-semver/issues/339
        ("rc=rc", "7.1.0-beta.1+build.1", "7.1.0-beta.2", {}, ok),
        # dev
        ("dev", "3.4.5", "3.4.5+dev.1", {}, ok),
        # custom build string once set, cannot be reset,
        # that's a python-semver bug https://github.com/python-semver/python-semver/issues/339
        (sep(("build", "dev")), "3.4.5", "3.4.5+build.2", {}, ok),
        (sep(("build", "dev", "dev")), "3.4.5", "3.4.5+build.3", {}, ok),
        (sep(("dev", "build=beta")), "3.4.5", "3.4.5+dev.2", {}, ok),
        # malformed specific version
        ("a", "1.0.0", "", {}, invalid),
        # malformed original version
        ("patch", "1.0.0b2", "", {}, invalid),
        # cannot separate with ", "
        (sep(("major", " minor")), "0.4.0", "1.1.0", {}, invalid),
    ),
)
def test_semver_scheme(
    isolation: Generator[Path, None, None],
    instructions: Optional[str],
    original: str,
    expected: str,
    settings: Mapping[str, bool],
    exp_error: Union[no_error, RaisesContext],
) -> None:
    scheme = SemverScheme(str(isolation), settings)
    with exp_error:
        assert scheme.update(instructions, original, {}) == expected



#!/usr/bin/env python


from datetime import date

__author__ = "Sven Siegmund"
__author_email__ = "sven.siegmund@gmail.com"
__maintainer__ = __author__
__maintainer_email__ = __author_email__
__release_date__ = date(year=2022, month=11, day=13)
__version__ = "0.2.4"



#!/usr/bin/env python

import os
from typing import Generator

import pytest
from hatch.config.constants import AppEnvVars, ConfigEnvVars, PublishEnvVars
from hatch.utils.fs import Path, temp_directory

collect_ignore = [
    # "test_semver_scheme.py",
    # "test_bump_instruction.py",
]


@pytest.fixture(scope="session", autouse=True)
def isolation() -> Generator[Path, None, None]:
    with temp_directory() as d:
        data_dir = d / "data"
        data_dir.mkdir()
        cache_dir = d / "cache"
        cache_dir.mkdir()

        default_env_vars = {
            AppEnvVars.NO_COLOR: "1",
            ConfigEnvVars.DATA: str(data_dir),
            ConfigEnvVars.CACHE: str(cache_dir),
            PublishEnvVars.REPO: "dev",
            "HATCH_SELF_TESTING": "true",
            "GIT_AUTHOR_NAME": "Foo Bar",
            "GIT_AUTHOR_EMAIL": "foo@bar.baz",
            "COLUMNS": "80",
            "LINES": "24",
        }
        with d.as_cwd(default_env_vars):
            os.environ.pop(AppEnvVars.ENV_ACTIVE, None)
            yield d



#!/usr/bin/env python


"""
Hatch-semver's own error types
"""


class HatchSemverError(Exception):
    """
    Base class for all hatch-semver errors.
    """

    pass


class ValidationError(HatchSemverError):
    """
    Raised if the new version is not a valid successor of the original version.
    """

    pass



#!/usr/bin/env python

"""
Implements the version scheme interface between hatch and python-semver.
"""

from copy import deepcopy
from operator import ge, gt
from typing import Mapping

from hatchling.version.scheme.plugin.interface import VersionSchemeInterface
from semver import VersionInfo

from .bump_instruction import BumpInstruction
from .errors import ValidationError


class SemverScheme(VersionSchemeInterface):
    """
    Implements the semantic version scheme.
    ### References
    - https://semver.org/
    - https://hatch.pypa.io/latest/plugins/version-scheme/reference/
    """

    PLUGIN_NAME = "semver"
    """
    The name of the plugin which is to be used in *pyproject.toml* in [`tool.hatch.version`].
    Value: `semver`
    """
    INSTRUCTION_SEPARATOR = ","
    """
    Separator of bump instructions which the user writes as an argument to `hatch version`.
    Value: `,` (comma)
    """

    def update(self, desired_version: str, original_version: str, version_data: Mapping) -> str:
        """
        Calculates the new version and returns it as a valid semver string.

        If the configuration option [`validate-bump`](https://hatch.pypa.io/latest/plugins/version-scheme/standard/#options) is *True* it calls
                [self.validate_bump](#hatch_semver.semver_scheme.SemverScheme.validate_bump)
                to check if the new version is valid
                as a successor of the original version.

        ### Parameters
        - *desired_version*: A series of commands carrying
                instructions how to bump the current version.
                These commands are separated by `SemverScheme.INSTRUCTION_SEPARATOR`.
                Each such command is in turn parsed and represented as an instance of
                `hatch_semver.bump_instruction.BumpInstruction`.
        - *original_version*: Project's original version. Must be a valid
                semantic version ([regex checker](https://regex101.com/r/Ly7O1x/3/)).
        - *version_data*: Poorly documented argument. Ignored entirely.
                But the plugin's interface requires it. If anyone knows what it
                could be used for, please let me know.

        ### Return
        Returns the new version as a valid semver string.
        """
        if not desired_version:
            return original_version
        original_version = VersionInfo.parse(original_version)
        current_version = deepcopy(original_version)
        instructions: str = desired_version
        validate = self.config.get("validate-bump", True)
        rc_bumps_patch = True
        for instruction in instructions.split(self.INSTRUCTION_SEPARATOR):
            bi = BumpInstruction(instruction)
            last_bump_was_build = False
            if bi.version_part == "build":
                current_version = current_version.bump_build(token=bi.token)
                last_bump_was_build = True
            elif bi.version_part == "release":
                current_version = current_version.finalize_version()
                rc_bumps_patch = True
            elif bi.is_specific:
                # Users may enter a specific version string to change nothing but
                # the metadata.
                # This would result in an unnecessary ValidationError if such an instruction
                # comes last in the instruction iterable.
                # To avoid this, we check if nothing but the metadata changed.
                # if so, we will pretend that last_bump_was_build
                temp_old_version = deepcopy(current_version)
                current_version = VersionInfo.parse(bi.version_part)
                if temp_old_version == current_version:
                    last_bump_was_build = True
                rc_bumps_patch = False
            else:
                if not rc_bumps_patch and bi.version_part == "prerelease":
                    current_version = current_version.bump_prerelease(token=bi.token)
                    rc_bumps_patch = True
                else:
                    current_version = current_version.next_version(
                        bi.version_part, prerelease_token=bi.token
                    )
                    rc_bumps_patch = False
        if validate:
            self.validate_bump(current_version, original_version, bumped_build=last_bump_was_build)
        return str(current_version)

    def validate_bump(
        self, current_version: VersionInfo, original_version: VersionInfo, bumped_build: bool
    ) -> None:
        """
        Validates if the current version is a valid successor of the original version.

        ### Parameters
        - *current_version*: the new version which is about to supersede the original version.
        - *original_version*: the original version which is about to be superseded by the new one.
        - *bumped_build*: information whether a bump of the build identifier was performed \
                as the last bump step.

        ### Raises
        Raises [ValidationError](../errors/#validationerror) if the *current_version* (new version) is not higher than the *original_version*. \
        In case only a build identifier bump was performed as the last bump, `ValidationError` is \
        raised if the new version is not at least of equal precedence.
        """
        if bumped_build:
            comparator = ge
            relation = "at least as high as"
        else:
            comparator = gt
            relation = "higher than"
        if comparator(current_version, original_version):
            return
        else:
            raise ValidationError(
                " ".join(
                    (
                        f"Version `{current_version}` is not {relation}",
                        f"the original version `{original_version}`",
                    )
                )
            )



#!/usr/bin/env python


from hatchling.plugin import hookimpl

from ..semver_scheme import SemverScheme

"""
The registration hook for [hatch](https://github.com/pypa/hatch/tree/master/backend/src/hatchling/plugin)
"""


@hookimpl
def hatch_register_version_scheme():
    """
    Returns `hatch_semver.semver_scheme.SemverScheme`.
    By this our plugin can be somehow registered in [hatch](https://hatch.pypa.io/).
    """
    return SemverScheme



#!/usr/bin/env python


from typing import Optional

import pytest

from hatch_semver.bump_instruction import BumpInstruction as BI


class TestInvalidToken:
    @pytest.mark.parametrize("part, token", (("alpha", "3"), ("beta", "2")))
    def test_invalid_prerelease_token(self, part: str, token: str) -> None:
        with pytest.raises(
            ValueError,
            match=f"{part} version cannot be set to {token} directly. Use 'prerelease={part}' instead",
        ):
            BI(f"{part}={token}")

    @pytest.mark.parametrize("part, token", (("dev", "3"),))
    def test_invalid_build_token(self, part: str, token: str) -> None:
        with pytest.raises(
            ValueError,
            match=f"{part} version cannot be set to {token} directly. Use 'build={part}' instead",
        ):
            BI(BI.sep.join((part, token)))

    @pytest.mark.parametrize(
        "part, norm_part, token",
        (
            ("major", "major", "3"),
            ("minor", "minor", "2"),
            ("patch", "patch", "4"),
            ("fix", "patch", "5"),
            ("micro", "patch", "6"),
            ("release", "release", "7"),
        ),
    )
    def test_invalid_part_token(self, part: str, norm_part: str, token: str) -> None:
        with pytest.raises(
            ValueError,
            match=f"{norm_part} version cannot be set to {token} specifically. Use {norm_part} alone",
        ):
            BI(BI.sep.join((part, token)))


class TestSeparatorNotInInstruction:
    @pytest.mark.parametrize(
        "instruction, part, token, specific",
        (
            ("", "", None, True),
            ("something-specific", "something-specific", None, True),
            ("release", "release", None, False),
            ("major", "major", None, False),
            ("minor", "minor", None, False),
            ("patch", "patch", None, False),
            ("fix", "patch", None, False),
            ("micro", "patch", None, False),
            ("pre", "prerelease", None, False),
            ("prerelease", "prerelease", None, False),
            ("pre-release", "prerelease", None, False),
            ("rc", "prerelease", None, False),
            ("build", "build", None, False),
            ("alpha", "prerelease", "alpha", False),
            ("beta", "prerelease", "beta", False),
            ("dev", "build", "dev", False),
        ),
    )
    def test_no_separator(
        self, instruction: str, part: str, token: Optional[str], specific: bool
    ) -> None:
        bi = BI(instruction)
        assert bi.version_part == part
        if token is None:
            assert bi.token is None
        else:
            assert bi.token == token
        assert bi.is_specific == specific


class TestSeparator:
    @pytest.mark.parametrize(
        "instruction",
        (
            BI.sep.join(("pre", "neat")),
            BI.sep.join(("build", "impossible", "token")),
            BI.sep.join(("pre-release", "even", "more", "insane", "token")),
        ),
    )
    def test_maxsplit(self, instruction: str) -> None:
        bi = BI(instruction)
        assert bi.token.count(BI.sep) == instruction.count(BI.sep) - 1
        assert not bi.is_specific

    @pytest.mark.parametrize(
        "part, norm_part, token, specific",
        (
            ("", "", "", True),
            ("pre", "prerelease", "mypre", False),
            ("prerelease", "prerelease", "some", False),
            ("pre-release", "prerelease", "the=hyphen-is=good", False),
            ("rc", "prerelease", "not;catching*invalid%tokens", False),
            ("build", "build", "1234", False),
            ("build", "build", "dev", False),
            ("build", "build", "devdrop", False),
        ),
    )
    def test_preserve_raw_part_and_raw_token(
        self, part: str, norm_part: str, token: str, specific: bool
    ) -> None:
        bi = BI(BI.sep.join((part, token)))
        assert bi.version_part is not None
        assert bi.token is not None
        assert bi.version_part == norm_part
        assert bi.token == token
        assert bi.is_specific == specific



#!/usr/bin/env python

from dataclasses import InitVar, dataclass, field
from typing import ClassVar, Optional


@dataclass
class BumpInstruction:
    """
    A dataclass for representing a single instruction from the
    chain of bump instructions (i.e. a part of the *desired_version*
    argument of `hatch_semver.semver_scheme.SemverScheme.update`.

    Takes a single string as an argument and calculates the values
    of an instance's attributes.
    """

    sep: ClassVar[str] = "="
    """
    Value: `=`.
    Separator of the version segment's name and the value of its identifier.
    Used to set a custom prerelease identifier, e.g. `prerelease=beta`
    or a custom build identifier, e.g. `build=develop`.
    """
    acceptable_version_parts: ClassVar[tuple[str]] = (
        "major",
        "minor",
        "patch",
        "prerelease",
        "build",
        "release",
    )
    """
    Used to decide whether the user wants to set a specific version or not
    (see [is_specific](#hatch_semver.bump_instruction.BumpInstruction.is_specific))
    If only a relative bump is requested, the bump instruction (which contains
    the name of the version part to bump) would correspond to one of the strings in
    this tuple. If not, then we assume that the string given in
    [version_part](#hatch_semver.bump_instruction.BumpInstruction.version_part)
    is a specific version.
    """
    version_part: str = field(init=False)
    """
    Is either one of the 
    [acceptable_version_parts](#hatch_semver.bump_instruction.BumpInstruction.acceptable_version_parts)
    or one of their aliases. Can also hold the value of the a specific version
    if the user wants to set it that way.
    """
    token: Optional[str] = field(init=False)
    """
    Holds the value of a custom identifier, e.g. `beta`, or `develop`.
    """
    is_specific: bool = field(init=False)
    """
    Information whether the user intends to set a specific version value,
    rather than bump any of the version segments.
    """
    instruction: InitVar[str]
    """
    Only available in the init phase of the instance. This is the raw
    unparsed bump instruction. One part of the string given to
    `hatch_semver.semver_scheme.SemverScheme.update`
    as the *desired_version* argument.
    """

    def __post_init__(self, instruction) -> None:
        if self.sep not in instruction:
            raw_version_part = instruction
            raw_token = None
        else:
            raw_version_part, raw_token = instruction.split(self.sep, maxsplit=1)
        self.version_part, self.token, self.is_specific = self.normalize_version_part(
            raw_version_part, raw_token
        )

    @classmethod
    def normalize_version_part(cls, part: str, token: str) -> tuple[str, str, bool]:
        """
        Effectively parses the raw string of the bump instruction into a tuple
        of atomic values which it returns. The `BumpInstruction.__post_init__`
        then assigns them to the instance's attributes.

        Version parts (version segments) can be referred to by many names.
        For instance the *patch* segment can be called *micro* or *fix*,
        the *pre-release* segment *pre*, *rc*, etc.

        This method normalizes all aliases to their native name
        (see [acceptable_version_parts](#hatch_semver.bump_instruction.BumpInstruction.acceptable_version_parts).
        It also handles shortcuts like `alpha` or `dev` which are actually
        `prerelease=alpha` or `build=dev` in their full syntax, i.e. they
        are to be understood as a version part with a custom identifier.

        ### Parameters
        - *cls*: the `BumpInstruction` class.
        - *part*: version part (`major`, `minor`, ...), alias (`rc`, `fix`, ...),
                a shortcut (`alpha`, `dev`, ...), or a specific version (e.g. `3.4.5`)
        - *token*: a custom token. If the user has specified something like `prerelease=alpha`,
                `alpha` would be the token (what the pre-release identifier should be called).

        ### Return
        a tuple of:

        - *part*: normalized to the native names defined in
                [acceptable_version_parts](#hatch_semver.bump_instruction.BumpInstruction.acceptable_version_parts), or a specific version
        - *token*: whatever the identifier's value should be
        - *specific*: information whether *part* contains a specific version,
                rather than a version part
        """
        if part in ("pre", "prerelease", "pre-release", "rc"):
            part = "prerelease"
        elif part in ("micro", "fix"):
            part = "patch"
        elif part in ("alpha", "beta"):
            if token:
                raise ValueError(
                    " ".join(
                        (
                            f"{part} version cannot be set to {token} directly.",
                            f"Use 'prerelease={part}' instead",
                        )
                    )
                )
            token = part
            part = "prerelease"
        elif part in ("dev",):
            if token:
                raise ValueError(
                    f"{part} version cannot be set to {token} directly. Use 'build={part}' instead"
                )
            token = part
            part = "build"
        if token:
            if part in ("major", "minor", "patch", "release"):
                raise ValueError(
                    f"{part} version cannot be set to {token} specifically. Use {part} alone"
                )
        if part in cls.acceptable_version_parts:
            specific = False
        else:
            specific = True
        return part, token, specific
